from __future__ import absolute_import
from __future__ import print_function
from datetime import tzinfo, timedelta, datetime
import tempfile as tf
from subprocess import Popen, PIPE
import re
import os
import sys

__author__ = "Steven Ring"

ZERO = timedelta(0)

# A UTC class.

class UTC(tzinfo):
    """
    All tidal predictions use UTC time zone
    """
    def utcoffset(self, dt):
        return ZERO
    def tzname(self, dt):
        return "UTC"
    def dst(self, dt):
        return ZERO

def unix_time(dt):
    epoch = datetime.utcfromtimestamp(0)
    delta = dt - epoch
    return delta.total_seconds()

class TimePoint(object):
    """
    Represents a place and a time 
    """
    def __init__(self, lon, lat, timestamp):
        """
        Construct a TimePoint object

        :param lon:
             The longitude of the point

        :param lat:
             The latitude of the point

        :param utc:
             A datetime instance specifying an 
             instant in time. If the timestamp is "naive"
             it will be interpreted as UTC.
        """

        self.lon = lon
        self.lat = lat
        if timestamp.tzinfo is None:
            self.timestamp = datetime.fromtimestamp(unix_time(timestamp), UTC())
        else:
            self.timestamp = timestamp

    def as_predict_input(self):
        """
        This timepoint as an input string to predict_tide
        """
        t = self.timestamp.astimezone(UTC())
        return "  %3.6f  %4.6f %6d %6d %6d %6d %6d %6d"  % (\
            self.lat, \
            self.lon, \
            t.year, \
            t.month, \
            t.day, \
            t.hour, \
            t.minute, \
            t.second)

    def __str__(self):
        return "At (%4.6f,%3.6f) %s"  % (\
            self.lon, \
            self.lat, \
            self.timestamp.isoformat())


class Tide(object):
    """
    Represents the tide height (and depth) at place and a time 
    """

    PATTERN = re.compile( \
        "\\s+" + \
        "(?P<lat>-?\\d+.\\d+)" + "\\s+" + \
        "(?P<lon>-?\\d+.\\d+)" + "\\s+" + \
        "(?P<mon>\\d{2})" + "\\." + \
        "(?P<day>\\d{2})" + "\\." + \
        "(?P<year>\\d{4})" + "\\s" + \
        "(?P<hour>\\d{2})" + ":" + \
        "(?P<min>\\d{2})" + ":" + \
        "(?P<sec>\\d{2})" + "\\s+" + \
        "(?P<z_m>-?\\d+.\\d+)" + "\\s+" + \
        "(?P<depth_m>-?\\d+.\\d+)") 

    @staticmethod
    def from_predict_output_line(text):
        """
        Parse a predict_tide output line and construct a Tide object
        """

        m = Tide.PATTERN.match(text.rstrip())
        if m is not None:
            return Tide( \
                TimePoint( \
                    float(m.group('lon')),
                    float(m.group('lat')),
                    datetime(
                        int(m.group('year')),
                        int(m.group('mon')),
                        int(m.group('day')),
                        int(m.group('hour')),
                        int(m.group('min')),
                        int(m.group('sec')))),
                float(m.group('z_m')),
                float(m.group('depth_m')))
        else:
            return None      

    def __init__(self, timepoint, tide_m, depth_m):
        """
        Construct a Tile object

        :param timepoint:
            A TimePoint instance specifying place and time

        :param tide_m:
            Predicted tide height in metres

        :param depth_m:
            Depth at the position according to Bathymetry 
            model
        """

        self.timepoint = timepoint
        self.tide_m = tide_m
        self.depth_m = depth_m

    def __str__(self):
        return "At (%4.6f,%3.6f) %s tide=%3.6f, depth=%3.6f"  % (\
            self.timepoint.lon, \
            self.timepoint.lat, \
            self.timepoint.timestamp.isoformat(), \
            self.tide_m, \
            self.depth_m)

def die(msg):
    print(msg, file=sys.stderr)
    sys.exit(1) 


def predict_tide(time_points, \
        model_path=os.getenv('OTPS_MODEL_ATLAS', '/var/share/TPX08_atlas_compact/Model_atlas')):
    """
    Predict the tide for the supplied timepoints
   
    :param time_points:
        A list of TimePoint instances giving the lat/lon/time of
        each required tide prediction

    :param model_path:
        The path to the OTPS Model_atlas file which describes the 
        model data files that are to be used for the tidal 
        prediction

    :return:
        a list of Tide instances containing the tidal predictions
    """

    # temporary files for input data to predict_tide

    data_file = tf.mkstemp()[1]
    out_file = tf.mkstemp()[1]

    # write output data

    with open(data_file, 'w') as data_f:
        for tp in time_points:
          data_f.write(tp.as_predict_input()+"\n")

    # run the predict_tide process

    cmd = ['predict_tide']
    stdin = model_path + "\n"
    stdin += data_file + "\n"
    stdin += "z\n"
    stdin += "\n"
    stdin += "AP\n"
    stdin += "oce\n"
    stdin += "1\n"
    stdin += out_file + "\n"
    try:
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    except Exception as  e:
        die("predict_tide command missing?: %s" % (str(e), ))

    (out_msg, err_msg) = p.communicate(input=stdin.encode())

    if p.wait() != 0:
        die(err_msg.rstrip())  # ensure operator sees result
    msg = out_msg.rstrip()

    # read and parse the output data
    tides = []
    with open(out_file, 'r') as result_file:
        for line in result_file:
            tide = Tide.from_predict_output_line(line)
            if tide is not None:
                tides.append(tide)

    # delete all files

    os.unlink(data_file)
    os.unlink(out_file)

    # return result

    return tides
